//---------------------------------------------------------------------------
#include <stdio.h>
#include "Allometry.h"
#include "ParsingFunctions.h"

////////////////////////////////////////////////////////////////////////////
// Constructor
// Edit history:
// -------------
// April 25, 2005 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
clAllometry::clAllometry()
{
  mp_AdultHeight = NULL;
  mp_AdultDiam = NULL;
  mp_SaplingHeight = NULL;
  mp_SaplingDiam = NULL;
  mp_SeedlingHeight = NULL;
  mp_SeedlingDiam = NULL;
  mp_fMaxTreeHeight = NULL;
  mp_fCanopyRadToDbh = NULL;
  mp_fCanopyRadExp = NULL;
  mp_fDbhToDiam10Ratio = NULL;
  mp_fFractOfHeightInCan = NULL;
  mp_fCanopyDepthExp = NULL;
  mp_fSlopeHeightDiam10 = NULL;
  mp_fSlopeAsympHeight = NULL;
  mp_fAdultLinearSlope = NULL;
  mp_fAdultLinearIntercept = NULL;
  mp_fAdultReverseLinearSlope = NULL;
  mp_fAdultReverseLinearIntercept = NULL;
  mp_fSaplingLinearSlope = NULL;
  mp_fSaplingLinearIntercept = NULL;
  mp_fSaplingReverseLinearSlope = NULL;
  mp_fSaplingReverseLinearIntercept = NULL;
  mp_fSeedlingLinearSlope = NULL;
  mp_fSeedlingLinearIntercept = NULL;
  mp_fSeedlingReverseLinearSlope = NULL;
  mp_fSeedlingReverseLinearIntercept = NULL;
}

////////////////////////////////////////////////////////////////////////////
// Destructor
// Edit history:
// -------------
// April 25, 2005 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
clAllometry::~clAllometry()
{
  delete[] mp_AdultHeight;
  delete[] mp_AdultDiam;
  delete[] mp_SaplingHeight;
  delete[] mp_SaplingDiam;
  delete[] mp_SeedlingHeight;
  delete[] mp_SeedlingDiam;
  delete[] mp_fMaxTreeHeight;
  delete[] mp_fCanopyRadToDbh;
  delete[] mp_fCanopyRadExp;
  delete[] mp_fDbhToDiam10Ratio;
  delete[] mp_fFractOfHeightInCan;
  delete[] mp_fCanopyDepthExp;
  delete[] mp_fSlopeHeightDiam10;
  delete[] mp_fSlopeAsympHeight;
  delete[] mp_fAdultLinearSlope;
  delete[] mp_fAdultLinearIntercept;
  delete[] mp_fAdultReverseLinearSlope;
  delete[] mp_fAdultReverseLinearIntercept;
  delete[] mp_fSaplingLinearSlope;
  delete[] mp_fSaplingLinearIntercept;
  delete[] mp_fSaplingReverseLinearSlope;
  delete[] mp_fSaplingReverseLinearIntercept;
  delete[] mp_fSeedlingLinearSlope;
  delete[] mp_fSeedlingLinearIntercept;
  delete[] mp_fSeedlingReverseLinearSlope;
  delete[] mp_fSeedlingReverseLinearIntercept;
}

////////////////////////////////////////////////////////////////////////////
// GetData()
// Edit history:
// -------------
// April 25, 2005 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
void clAllometry::GetData( DOMDocument * p_oDoc, clTreePopulation * p_oPop )
{
  try
  {
    DOMElement * p_oElement = p_oDoc->getDocumentElement();
    int * p_iWhatFunction; //for getting which function to use for each
    //species
    floatVal * p_fTemp = NULL; //for extracting values for a subset of all species
    int i, iHowMany;

    m_iNumSpecies = p_oPop->GetNumberOfSpecies();
    p_iWhatFunction = new int[m_iNumSpecies];

    //Declare our arrays
    mp_AdultHeight = new Ptr2Allometry[m_iNumSpecies];
    mp_AdultDiam = new Ptr2Allometry[m_iNumSpecies];
    mp_SaplingHeight = new Ptr2Allometry[m_iNumSpecies];
    mp_SaplingDiam = new Ptr2Allometry[m_iNumSpecies];
    mp_SeedlingHeight = new Ptr2Allometry[m_iNumSpecies];
    mp_SeedlingDiam = new Ptr2Allometry[m_iNumSpecies];
    mp_fMaxTreeHeight = new float[m_iNumSpecies];
    mp_fCanopyRadToDbh = new float[m_iNumSpecies];
    mp_fCanopyRadExp = new float[m_iNumSpecies];
    mp_fDbhToDiam10Ratio = new float[m_iNumSpecies];
    mp_fFractOfHeightInCan = new float[m_iNumSpecies];
    mp_fCanopyDepthExp = new float[m_iNumSpecies];
    mp_fSlopeHeightDiam10 = new float[m_iNumSpecies];
    mp_fSlopeAsympHeight = new float[m_iNumSpecies];
    mp_fAdultLinearSlope = new float[m_iNumSpecies];
    mp_fAdultLinearIntercept = new float[m_iNumSpecies];
    mp_fAdultReverseLinearSlope = new float[m_iNumSpecies];
    mp_fAdultReverseLinearIntercept = new float[m_iNumSpecies];
    mp_fSaplingLinearSlope = new float[m_iNumSpecies];
    mp_fSaplingLinearIntercept = new float[m_iNumSpecies];
    mp_fSaplingReverseLinearSlope = new float[m_iNumSpecies];
    mp_fSaplingReverseLinearIntercept = new float[m_iNumSpecies];
    mp_fSeedlingLinearSlope = new float[m_iNumSpecies];
    mp_fSeedlingLinearIntercept = new float[m_iNumSpecies];
    mp_fSeedlingReverseLinearSlope = new float[m_iNumSpecies];
    mp_fSeedlingReverseLinearIntercept = new float[m_iNumSpecies];

    //Get all-species parameters
    FillSpeciesSpecificValue( p_oElement, "tr_canopyHeight", "tr_chVal", mp_fMaxTreeHeight, p_oPop, true );
    FillSpeciesSpecificValue( p_oElement, "tr_canopyRadiusToDBH", "tr_crtdVal", mp_fCanopyRadToDbh, p_oPop, true );
    FillSpeciesSpecificValue( p_oElement, "tr_canopyRadiusExponent", "tr_creVal", mp_fCanopyRadExp, p_oPop, true );
    FillSpeciesSpecificValue( p_oElement, "tr_fractOfHeightInCanopy", "tr_fohicVal", mp_fFractOfHeightInCan, p_oPop, true );
    FillSpeciesSpecificValue( p_oElement, "tr_canopyDepthExponent", "tr_cdeVal", mp_fCanopyDepthExp, p_oPop, true );
    FillSpeciesSpecificValue( p_oElement, "tr_conversionDiam10ToDBH", "tr_cdtdVal", mp_fDbhToDiam10Ratio, p_oPop, true );

    //If the value of the dbh to diam10 conversion factor is zero, throw error
    for ( int i = 0; i < m_iNumSpecies; i++ )
    {
      if ( mp_fDbhToDiam10Ratio[i] == 0 )
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clAllometry::GetData" );
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cMoreInfo, "Diameter at 10 cm to DBH conversion may not equal 0." );
        throw( stcErr );
      }
    }

    //Start by getting the allometry function for each species for each life
    //history stage
    //***********************
    // Adults
    //***********************
    FillSpeciesSpecificValue( p_oElement, "tr_whatAdultHeightDiam", "tr_wahdVal", p_iWhatFunction, p_oPop, true );

    //Use the values to set function pointers - if any are not a valid function
    //enum number, throw an error
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( standard == p_iWhatFunction[i] )
      {
        mp_AdultHeight[i] = & clAllometry::CalcStandardSapAdultHeight;
        mp_AdultDiam[i] = & clAllometry::CalcStandardSapAdultDbh;
      }
      else if ( linear == p_iWhatFunction[i] )
      {
        mp_AdultHeight[i] = & clAllometry::CalcLinearAdultHeight;
        mp_AdultDiam[i] = & clAllometry::CalcLinearAdultDbh;
      }
      else if ( reverse_linear == p_iWhatFunction[i] )
      {
        mp_AdultHeight[i] = & clAllometry::CalcReverseLinearAdultHeight;
        mp_AdultDiam[i] = & clAllometry::CalcReverseLinearAdultDbh;
      }
      else
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clAllometry::GetData" );
        sprintf( stcErr.cMoreInfo, "%s%d", "Invalid adult height-diam function code: ", p_iWhatFunction[i] );
        throw( stcErr );
      }
    }

    //If any of the functions are standard, read standard parameters
    iHowMany = 0;
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( standard == p_iWhatFunction[i] ) iHowMany++;
    }
    if ( iHowMany > 0 )
    {
      if ( p_fTemp )
      {
        delete[] p_fTemp; p_fTemp = NULL;
      }
      p_fTemp = new floatVal[iHowMany];
      iHowMany = 0;
      for ( i = 0; i < m_iNumSpecies; i++ )
      {
        if ( standard == p_iWhatFunction[i] )
        {
          p_fTemp[iHowMany].code = i;
          iHowMany++;
        }
      }

      FillSpeciesSpecificValue( p_oElement, "tr_slopeOfAsymHeight", "tr_soahVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        mp_fSlopeAsympHeight[p_fTemp[i].code] = p_fTemp[i].val;
      }
    }

    //If any of the functions are linear, read linear parameters
    iHowMany = 0;
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( linear == p_iWhatFunction[i] ) iHowMany++;
    }
    if ( iHowMany > 0 )
    {
      if ( p_fTemp )
      {
        delete[] p_fTemp; p_fTemp = NULL;
      }
      p_fTemp = new floatVal[iHowMany];
      iHowMany = 0;
      for ( i = 0; i < m_iNumSpecies; i++ )
      {
        if ( linear == p_iWhatFunction[i] )
        {
          p_fTemp[iHowMany].code = i;
          iHowMany++;
        }
      }

      //Linear adult slope - throw an error if any values are 0
      FillSpeciesSpecificValue( p_oElement, "tr_adultLinearSlope", "tr_alsVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        if ( 0 == p_fTemp[i].val )
        {
          modelErr stcErr;
          stcErr.iErrorCode = BAD_DATA;
          strcpy( stcErr.cFunction, "clAllometry::GetData" );
          strcpy( stcErr.cMoreInfo, "No values in adult linear slope may be 0." );
          throw( stcErr );

        }
        mp_fAdultLinearSlope[p_fTemp[i].code] = p_fTemp[i].val;
      }

      //Linear adult intercept
      FillSpeciesSpecificValue( p_oElement, "tr_adultLinearIntercept", "tr_aliVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        mp_fAdultLinearIntercept[p_fTemp[i].code] = p_fTemp[i].val;
      }
    }

    //If any of the functions are reverse linear, read linear parameters
    iHowMany = 0;
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( reverse_linear == p_iWhatFunction[i] ) iHowMany++;
    }
    if ( iHowMany > 0 )
    {
      if ( p_fTemp )
      {
        delete[] p_fTemp; p_fTemp = NULL;
      }
      p_fTemp = new floatVal[iHowMany];
      iHowMany = 0;
      for ( i = 0; i < m_iNumSpecies; i++ )
      {
        if ( reverse_linear == p_iWhatFunction[i] )
        {
          p_fTemp[iHowMany].code = i;
          iHowMany++;
        }
      }

      //Reverse linear adult slope - throw an error if any are 0
      FillSpeciesSpecificValue( p_oElement, "tr_adultReverseLinearSlope", "tr_arlsVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        if ( 0 == p_fTemp[i].val )
        {
          modelErr stcErr;
          stcErr.iErrorCode = BAD_DATA;
          strcpy( stcErr.cFunction, "clAllometry::GetData" );
          strcpy( stcErr.cMoreInfo, "No values in adult reverse linear slope may be 0." );
          throw( stcErr );

        }
        mp_fAdultReverseLinearSlope[p_fTemp[i].code] = p_fTemp[i].val;
      }

      //Reverse linear adult intercept
      FillSpeciesSpecificValue( p_oElement, "tr_adultReverseLinearIntercept", "tr_arliVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        mp_fAdultReverseLinearIntercept[p_fTemp[i].code] = p_fTemp[i].val;
      }
    }

    //***********************
    // Saplings
    //***********************
    FillSpeciesSpecificValue( p_oElement, "tr_whatSaplingHeightDiam", "tr_wsahdVal", p_iWhatFunction, p_oPop, true );

    //Use the values to set function pointers - if any are not a valid function
    //enum number, throw an error
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( standard == p_iWhatFunction[i] )
      {
        mp_SaplingHeight[i] = & clAllometry::CalcStandardSapAdultHeight;
        mp_SaplingDiam[i] = & clAllometry::CalcStandardSapAdultDbh;
      }
      else if ( linear == p_iWhatFunction[i] )
      {
        mp_SaplingHeight[i] = & clAllometry::CalcLinearSaplingHeight;
        mp_SaplingDiam[i] = & clAllometry::CalcLinearSaplingDbh;
      }
      else if ( reverse_linear == p_iWhatFunction[i] )
      {
        mp_SaplingHeight[i] = & clAllometry::CalcReverseLinearSaplingHeight;
        mp_SaplingDiam[i] = & clAllometry::CalcReverseLinearSaplingDbh;
      }
      else
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clAllometry::GetData" );
        sprintf( stcErr.cMoreInfo, "%s%d", "Invalid sapling height-diam function code: ", p_iWhatFunction[i] );
        throw( stcErr );
      }
    }

    //If any of the functions are standard, read standard parameters
    iHowMany = 0;
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( standard == p_iWhatFunction[i] ) iHowMany++;
    }
    if ( iHowMany > 0 )
    {
      if ( p_fTemp )
      {
        delete[] p_fTemp; p_fTemp = NULL;
      }
      p_fTemp = new floatVal[iHowMany];
      iHowMany = 0;
      for ( i = 0; i < m_iNumSpecies; i++ )
      {
        if ( standard == p_iWhatFunction[i] )
        {
          p_fTemp[iHowMany].code = i;
          iHowMany++;
        }
      }

      FillSpeciesSpecificValue( p_oElement, "tr_slopeOfAsymHeight", "tr_soahVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        mp_fSlopeAsympHeight[p_fTemp[i].code] = p_fTemp[i].val;
      }
    }

    //If any of the functions are linear, read linear parameters
    iHowMany = 0;
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( linear == p_iWhatFunction[i] ) iHowMany++;
    }
    if ( iHowMany > 0 )
    {
      if ( p_fTemp )
      {
        delete[] p_fTemp; p_fTemp = NULL;
      }
      p_fTemp = new floatVal[iHowMany];
      iHowMany = 0;
      for ( i = 0; i < m_iNumSpecies; i++ )
      {
        if ( linear == p_iWhatFunction[i] )
        {
          p_fTemp[iHowMany].code = i;
          iHowMany++;
        }
      }

      //Linear sapling slope - throw an error if any values are 0
      FillSpeciesSpecificValue( p_oElement, "tr_saplingLinearSlope", "tr_salsVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        if ( 0 == p_fTemp[i].val )
        {
          modelErr stcErr;
          stcErr.iErrorCode = BAD_DATA;
          strcpy( stcErr.cFunction, "clAllometry::GetData" );
          strcpy( stcErr.cMoreInfo, "No values in sapling linear slope may be 0." );
          throw( stcErr );

        }
        mp_fSaplingLinearSlope[p_fTemp[i].code] = p_fTemp[i].val;
      }

      //Linear sapling intercept
      FillSpeciesSpecificValue( p_oElement, "tr_saplingLinearIntercept", "tr_saliVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        mp_fSaplingLinearIntercept[p_fTemp[i].code] = p_fTemp[i].val;
      }
    }

    //If any of the functions are reverse linear, read linear parameters
    iHowMany = 0;
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( reverse_linear == p_iWhatFunction[i] ) iHowMany++;
    }
    if ( iHowMany > 0 )
    {
      if ( p_fTemp )
      {
        delete[] p_fTemp; p_fTemp = NULL;
      }
      p_fTemp = new floatVal[iHowMany];
      iHowMany = 0;
      for ( i = 0; i < m_iNumSpecies; i++ )
      {
        if ( reverse_linear == p_iWhatFunction[i] )
        {
          p_fTemp[iHowMany].code = i;
          iHowMany++;
        }
      }

      //Reverse linear sapling slope - throw an error if any values are 0
      FillSpeciesSpecificValue( p_oElement, "tr_saplingReverseLinearSlope", "tr_sarlsVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        if ( 0 == p_fTemp[i].val )
        {
          modelErr stcErr;
          stcErr.iErrorCode = BAD_DATA;
          strcpy( stcErr.cFunction, "clAllometry::GetData" );
          strcpy( stcErr.cMoreInfo, "No values in sapling reverse linear slope may be 0." );
          throw( stcErr );

        }
        mp_fSaplingReverseLinearSlope[p_fTemp[i].code] = p_fTemp[i].val;
      }

      //Reverse linear sapling intercept
      FillSpeciesSpecificValue( p_oElement, "tr_saplingReverseLinearIntercept", "tr_sarliVal", p_fTemp,
           iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        mp_fSaplingReverseLinearIntercept[p_fTemp[i].code] = p_fTemp[i].val;
      }
    }

    //***********************
    // Seedlings
    //***********************
    FillSpeciesSpecificValue( p_oElement, "tr_whatSeedlingHeightDiam", "tr_wsehdVal", p_iWhatFunction, p_oPop, true );

    //Use the values to set function pointers - if any are not a valid function
    //enum number, throw an error
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( standard == p_iWhatFunction[i] )
      {
        mp_SeedlingHeight[i] = & clAllometry::CalcStandardSeedlingHeight;
        mp_SeedlingDiam[i] = & clAllometry::CalcStandardSeedlingDiam10;
      }
      else if ( linear == p_iWhatFunction[i] )
      {
        mp_SeedlingHeight[i] = & clAllometry::CalcLinearSeedlingHeight;
        mp_SeedlingDiam[i] = & clAllometry::CalcLinearSeedlingDiam10;
      }
      else if ( reverse_linear == p_iWhatFunction[i] )
      {
        mp_SeedlingHeight[i] = & clAllometry::CalcReverseLinearSeedlingHeight;
        mp_SeedlingDiam[i] = & clAllometry::CalcReverseLinearSeedlingDiam10;
      }
      else
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clAllometry::GetData" );
        sprintf( stcErr.cMoreInfo, "%s%d", "Invalid seedling height-diam function code: ", p_iWhatFunction[i] );
        throw( stcErr );
      }
    }

    //If any of the functions are standard, read standard parameters
    iHowMany = 0;
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( standard == p_iWhatFunction[i] ) iHowMany++;
    }
    if ( iHowMany > 0 )
    {
      if ( p_fTemp )
      {
        delete[] p_fTemp; p_fTemp = NULL;
      }
      p_fTemp = new floatVal[iHowMany];
      iHowMany = 0;
      for ( i = 0; i < m_iNumSpecies; i++ )
      {
        if ( standard == p_iWhatFunction[i] )
        {
          p_fTemp[iHowMany].code = i;
          iHowMany++;
        }
      }

      FillSpeciesSpecificValue( p_oElement, "tr_slopeOfHeight-Diam10", "tr_sohdVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        mp_fSlopeHeightDiam10[p_fTemp[i].code] = p_fTemp[i].val;
      }
    }

    //If any of the functions are linear, read linear parameters
    iHowMany = 0;
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( linear == p_iWhatFunction[i] ) iHowMany++;
    }
    if ( iHowMany > 0 )
    {
      if ( p_fTemp )
      {
        delete[] p_fTemp; p_fTemp = NULL;
      }
      p_fTemp = new floatVal[iHowMany];
      iHowMany = 0;
      for ( i = 0; i < m_iNumSpecies; i++ )
      {
        if ( linear == p_iWhatFunction[i] )
        {
          p_fTemp[iHowMany].code = i;
          iHowMany++;
        }
      }

      //Linear seedling slope - throw error if any values are 0
      FillSpeciesSpecificValue( p_oElement, "tr_seedlingLinearSlope", "tr_selsVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        if ( 0 == p_fTemp[i].val )
        {
          modelErr stcErr;
          stcErr.iErrorCode = BAD_DATA;
          strcpy( stcErr.cFunction, "clAllometry::GetData" );
          strcpy( stcErr.cMoreInfo, "No values in seedling linear slope may be 0." );
          throw( stcErr );

        }
        mp_fSeedlingLinearSlope[p_fTemp[i].code] = p_fTemp[i].val;
      }

      //Linear seedling intercept
      FillSpeciesSpecificValue( p_oElement, "tr_seedlingLinearIntercept", "tr_seliVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        mp_fSeedlingLinearIntercept[p_fTemp[i].code] = p_fTemp[i].val;
      }
    }

    //If any of the functions are reverse linear, read linear parameters
    iHowMany = 0;
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      if ( reverse_linear == p_iWhatFunction[i] ) iHowMany++;
    }
    if ( iHowMany > 0 )
    {
      if ( p_fTemp )
      {
        delete[] p_fTemp; p_fTemp = NULL;
      }
      p_fTemp = new floatVal[iHowMany];
      iHowMany = 0;
      for ( i = 0; i < m_iNumSpecies; i++ )
      {
        if ( reverse_linear == p_iWhatFunction[i] )
        {
          p_fTemp[iHowMany].code = i;
          iHowMany++;
        }
      }

      //Reverse linear seedling slope - throw an error if any are 0
      FillSpeciesSpecificValue( p_oElement, "tr_seedlingReverseLinearSlope", "tr_serlsVal", p_fTemp, iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        if ( 0 == p_fTemp[i].val )
        {
          modelErr stcErr;
          stcErr.iErrorCode = BAD_DATA;
          strcpy( stcErr.cFunction, "clAllometry::GetData" );
          strcpy( stcErr.cMoreInfo, "No values in seedling reverse linear slope may be 0." );
          throw( stcErr );
        }
        mp_fSeedlingReverseLinearSlope[p_fTemp[i].code] = p_fTemp[i].val;
      }

      //Reverse linear seedling intercept
      FillSpeciesSpecificValue( p_oElement, "tr_seedlingReverseLinearIntercept", "tr_serliVal", p_fTemp,
           iHowMany, p_oPop, true );
      for ( i = 0; i < iHowMany; i++ )
      {
        mp_fSeedlingReverseLinearIntercept[p_fTemp[i].code] = p_fTemp[i].val;
      }
    }

    delete[] p_iWhatFunction;
    delete[] p_fTemp;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clAllometry::GetData" );
    throw( stcErr );
  }
}

////////////////////////////////////////////////////////////////////////////
// CalcAdultHeight()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// April 25, 2005 - Rewrote to use function pointers (LEM)
////////////////////////////////////////////////////////////////////////////
float clAllometry::CalcAdultHeight( const float & fDbh, const int & iSpecies )
{
  if ( iSpecies >= m_iNumSpecies || iSpecies < 0 )
  {
    //invalid species - throw an error
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::CalcAdultHeight" );
    sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized species ", iSpecies );
    throw( stcErr );
  }
  return (* this.*mp_AdultHeight[iSpecies] ) ( fDbh, iSpecies );
}

////////////////////////////////////////////////////////////////////////////
// CalcAdultDbh()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// April 25, 2005 - Rewrote to use function pointers (LEM)
////////////////////////////////////////////////////////////////////////////
float clAllometry::CalcAdultDbh( float fHeight, const int & iSpecies )
{
  if ( iSpecies >= m_iNumSpecies || iSpecies < 0 )
  {
    //invalid species - throw an error
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::CalcAdultDbh" );
    sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized species ", iSpecies );
    throw( stcErr );
  }
  //If the height is greater than or equal to canopy height, throw bad data error
  if ( fHeight >= mp_fMaxTreeHeight[iSpecies] )
  {
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::CalcAdultDbh" );
    sprintf( stcErr.cMoreInfo, "%s%s", "Cannot set tree height greater than ", "maximum canopy height for this species." );
    throw( stcErr );
  }
  return (* this.*mp_AdultDiam[iSpecies] ) ( fHeight, iSpecies );
}

////////////////////////////////////////////////////////////////////////////
// CalcSaplingHeight()
// Edit history:
// -------------
// April 25, 2005 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
float clAllometry::CalcSaplingHeight( const float & fDbh, const int & iSpecies )
{
  if ( iSpecies >= m_iNumSpecies || iSpecies < 0 )
  {
    //invalid species - throw an error
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::CalcSaplingHeight" );
    sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized species ", iSpecies );
    throw( stcErr );
  }
  return (* this.*mp_SaplingHeight[iSpecies] ) ( fDbh, iSpecies );
}

////////////////////////////////////////////////////////////////////////////
// CalcSaplingDbh()
// Edit history:
// -------------
// April 25, 2005 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
float clAllometry::CalcSaplingDbh( float fHeight, const int & iSpecies )
{
  if ( iSpecies >= m_iNumSpecies || iSpecies < 0 )
  {
    //invalid species - throw an error
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::CalcSaplingDbh" );
    sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized species ", iSpecies );
    throw( stcErr );
  }
  //If the height is greater than or equal to canopy height, throw bad data error
  if ( fHeight >= mp_fMaxTreeHeight[iSpecies] )
  {
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::CalcSaplingDbh" );
    sprintf( stcErr.cMoreInfo, "%s%s", "Cannot set tree height greater than ", "maximum canopy height for this species." );
    throw( stcErr );
  }
  return (* this.*mp_SaplingDiam[iSpecies] ) ( fHeight, iSpecies );
}

////////////////////////////////////////////////////////////////////////////
// CalcAdultCanRadius()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
////////////////////////////////////////////////////////////////////////////
float clAllometry::CalcAdultCanRadius( const float & fDbh, const int & iSpecies )
{
  float fCanrad;
  if ( iSpecies >= m_iNumSpecies || iSpecies < 0 )
  {
    //invalid species - throw an error
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::CalcAdultCanRadius" );
    sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized species ", iSpecies );
    throw( stcErr );
  }
  fCanrad = mp_fCanopyRadToDbh[iSpecies] * pow( fDbh, mp_fCanopyRadExp[iSpecies] );
  if ( fCanrad > MAX_CANOPY_RADIUS )
    fCanrad = MAX_CANOPY_RADIUS;
  return fCanrad;
}

////////////////////////////////////////////////////////////////////////////
// CalcAdultCanDepth()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
////////////////////////////////////////////////////////////////////////////
float clAllometry::CalcAdultCanDepth( float fHeight, const int & iSpecies )
{
  if ( iSpecies >= m_iNumSpecies || iSpecies < 0 )
  {
    //invalid species - throw an error
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::CalcAdultCanDepth" );
    sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized species ", iSpecies );
    throw( stcErr );
  }
  return mp_fFractOfHeightInCan[iSpecies] * pow( fHeight, mp_fCanopyDepthExp[iSpecies] );
}

////////////////////////////////////////////////////////////////////////////
// CalcSeedlingHeight()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// April 25, 2005 - Rewrote to use function pointers (LEM)
////////////////////////////////////////////////////////////////////////////
float clAllometry::CalcSeedlingHeight( float fDiam10, const int & iSpecies )
{
  if ( iSpecies >= m_iNumSpecies || iSpecies < 0 )
  {
    //invalid species - throw an error
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::CalcSeedlingHeight" );
    sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized species ", iSpecies );
    throw( stcErr );
  }
  return (* this.*mp_SeedlingHeight[iSpecies] ) ( fDiam10, iSpecies );
}

////////////////////////////////////////////////////////////////////////////
// CalcSeedlingDiam10()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// April 25, 2005 - Rewrote to use function pointers (LEM)
////////////////////////////////////////////////////////////////////////////
float clAllometry::CalcSeedlingDiam10( float fHeight, const int & iSpecies )
{
  if ( iSpecies >= m_iNumSpecies || iSpecies < 0 )
  {
    //invalid species - throw an error
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::CalcSeedlingDiam10" );
    sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized species ", iSpecies );
    throw( stcErr );
  }
  //If the height is greater than or equal to 30.1, throw bad data error
  if ( fHeight >= 30.1 )
  {
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::CalcSeedlingDiam10" );
    sprintf( stcErr.cMoreInfo, "%s%s", "Cannot set tree height greater than ", "maximum tree height for this species." );
    throw( stcErr );
  }
  return (* this.*mp_SeedlingDiam[iSpecies] ) ( fHeight, iSpecies );
}

////////////////////////////////////////////////////////////////////////////
// ConvertDiam10ToDbh()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
////////////////////////////////////////////////////////////////////////////
float clAllometry::ConvertDiam10ToDbh( float fDiam10, const int & iSpecies )
{
  if ( iSpecies >= m_iNumSpecies || iSpecies < 0 )
  {
    //invalid species - throw an error
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::ConvertDiam10ToDbh" );
    sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized species ", iSpecies );
    throw( stcErr );
  }
  return fDiam10 * mp_fDbhToDiam10Ratio[iSpecies];
}

////////////////////////////////////////////////////////////////////////////
// ConvertDbhToDiam10()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
////////////////////////////////////////////////////////////////////////////
float clAllometry::ConvertDbhToDiam10( const float & fDbh, const int & iSpecies )
{
  if ( iSpecies >= m_iNumSpecies || iSpecies < 0 )
  {
    //invalid species - throw an error
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clAllometry::ConvertDbhToDiam10" );
    sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized species ", iSpecies );
    throw( stcErr );
  }
  return fDbh / mp_fDbhToDiam10Ratio[iSpecies];
}

//////////////////////////////////////////////////////////////////////////////
// GetMaxTreeHeight()
// Edit history :
// -------------
// April 25, 2005 - Moved to here from clTreePopulation (LEM )
//////////////////////////////////////////////////////////////////////////////
float clAllometry::GetMaxTreeHeight( int iSpecies )
{
  try
  {
    if ( iSpecies < 0 || iSpecies >= m_iNumSpecies )
    {
      //invalid species - throw an error
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clTreePopulation::GetMaxTreeHeight" );
      sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized species ", iSpecies );
      throw( stcErr );
    }
    return mp_fMaxTreeHeight[iSpecies];
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clTreePopulation::GetMaxTreeHeight" );
    throw( stcErr );
  }
}
