//---------------------------------------------------------------------------
#include "RelativeGrowth.h"
#include "GrowthOrg.h"
#include "Tree.h"
#include "SimManager.h"
#include <math.h>
#include <stdio.h>
//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// SetNameData()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
//////////////////////////////////////////////////////////////////////////////
void clRelativeGrowth::SetNameData( char * cNameString )
{
  try
  {

    //Check the string passed and set the flags accordingly
    if ( strcmp( "relradialgrowth", cNameString ) == 0 )
    {
      m_bConstRadialLimited = true;
      m_bConstBasalAreaLimited = false;
      m_iGrowthMethod = diameter_auto;
    }
    else if ( strcmp( "relbagrowth", cNameString ) == 0 )
    {
      m_bConstRadialLimited = false;
      m_bConstBasalAreaLimited = true;
      m_iGrowthMethod = diameter_auto;
    }
    else if ( strcmp( "relunlimgrowth", cNameString ) == 0 )
    {
      m_bConstRadialLimited = false;
      m_bConstBasalAreaLimited = false;
      m_iGrowthMethod = diameter_auto;
    }
    else if ( strcmp( "relradialgrowth diam only", cNameString ) == 0 )
    {
      m_bConstRadialLimited = true;
      m_bConstBasalAreaLimited = false;
      m_iGrowthMethod = diameter_only;
    }
    else if ( strcmp( "relbagrowth diam only", cNameString ) == 0 )
    {
      m_bConstRadialLimited = false;
      m_bConstBasalAreaLimited = true;
      m_iGrowthMethod = diameter_only;
    }
    else if ( strcmp( "relunlimgrowth diam only", cNameString ) == 0 )
    {
      m_bConstRadialLimited = false;
      m_bConstBasalAreaLimited = false;
      m_iGrowthMethod = diameter_only;
    }
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clRelativeGrowth::SetNameData" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// DoShellSetup()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// August 27, 2004 - Made light required (LEM)
//////////////////////////////////////////////////////////////////////////////
void clRelativeGrowth::DoShellSetup( DOMDocument * p_oDoc )
{
  try
  {
    short int iNumSpecies = mp_oGrowthOrg->GetNumberOfSpecies();

    m_fNumberYearsPerTimestep = mp_oSimManager->GetNumberOfYearsPerTimestep();

    //Declare the arrays we'd like read
    mp_fAsympDiamGrowth = new float[iNumSpecies];
    mp_fSlopeGrowthResponse = new float[iNumSpecies];

    if ( m_bConstRadialLimited )
    {
      mp_fAdultConstRadInc = new float[iNumSpecies];
    }

    if ( m_bConstBasalAreaLimited )
    {
      mp_fAdultConstBAInc = new float[iNumSpecies];
    }

    //Read the base variables
    GetParameterFileData( p_oDoc );

    //Make sure all species/type combos have "Light" registered
    for ( int i = 0; i < m_iNumSpeciesTypeCombos; i++ )
    {
      if ( -1 == mp_oGrowthOrg->GetLightCode(mp_whatSpeciesTypeCombos[i].iSpecies,
                                           mp_whatSpeciesTypeCombos[i].iType ) )
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clRelativeGrowth::DoShellSetup" );
        sprintf( stcErr.cMoreInfo, "%s%d%s%d%s%s",
                 "Type/species combo species=",
                 mp_whatSpeciesTypeCombos[i].iSpecies,
                 " type=", mp_whatSpeciesTypeCombos[i].iType,
                 " does not have a required light behavior." );
        stcErr.iErrorCode = BAD_DATA;
        throw( stcErr );
      }
    }
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clRelativeGrowth::DoShellSetup" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// CalcDiameterGrowthValue()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
//////////////////////////////////////////////////////////////////////////////
float clRelativeGrowth::CalcDiameterGrowthValue( clTree * p_oTree, clTreePopulation * p_oPop, float fHeightGrowth )
{
  try
  {
    float fAnnualRelativeGrowth, //amount of annual relative growth
         fGli, //tree's gli value
         fDiam, //tree's diameter value
         fCompoundRelativeGrowth, //relative growth compounded to a complete
         //timestep
         fAmountDiamIncrease; //amount by which the tree's diameter will
    //increase
    short int iSpecies = p_oTree->GetSpecies(), iType = p_oTree->GetType(),
         iLightCode = mp_oGrowthOrg->GetLightCode( iSpecies, iType );

    //***************************************
    // Calculate the amount of growth
    //***************************************

    //Get the gli value for this tree
    p_oTree->GetValue( iLightCode, & fGli );

    //Get the appropriate diameter for this tree
    p_oTree->GetValue( mp_oGrowthOrg->GetDiamCode( iSpecies, iType ), & fDiam );

    //Get the value of the Michaelis-Menton function
    fAnnualRelativeGrowth = CalculateMichaelisMentonFunction( iSpecies, fGli );

    //Compound the relative growth over the number of years/time step
    fCompoundRelativeGrowth = pow( 1.0 + fAnnualRelativeGrowth, m_fNumberYearsPerTimestep );

    //Calculate amount of diameter increase based on the tree's diameter
    fAmountDiamIncrease = fDiam * ( fCompoundRelativeGrowth - 1.0 );

    fAmountDiamIncrease = ApplyGrowthLimits( iSpecies, fAmountDiamIncrease, fDiam );

    return fAmountDiamIncrease;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clRelativeGrowth::CalcGrowthValue" );
    throw( stcErr );
  }
}
//---------------------------------------------------------------------------
